#!/usr/bin/env python3
""" libgh - GitHub scraping library and tool
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging

from bs4 import BeautifulSoup

from .common import GITHUB_URL, CACHE_DIR, REQUESTS_MIN_DELAY, REQUESTS_PER_HOUR
from .libpnu2 import get_url_data

####################################################################################################
def load_repository(account_name, repository_name, cache_days, force_fetch=False):
    """ Returns a dictionary of repository information """
    repository = {}

    try:
        data, response = get_url_data(
            f"{GITHUB_URL}/{account_name}/{repository_name}",
            CACHE_DIR,
            cache_days=cache_days,
            cache_index=True,
            force_fetch=force_fetch,
            min_delay=REQUESTS_MIN_DELAY,
            max_per_hour=REQUESTS_PER_HOUR
        )
    except (LookupError, PermissionError) as error:
        logging.error("libgh: %s", error)
        return repository
    for item in response:
        if item[0].startswith("x-ratelimit"):
            logging.debug("libgh: HTTP response: %s=%s", item[0], item[1])

    soup = BeautifulSoup(data, "html.parser")

    #html = soup.select_one('[itemprop="author"]')
    #html = soup.select_one('[itemprop="name"]')

    # Forked from (if relevant)
    html = soup.select_one('[data-hovercard-type="repository"]')
    if html is not None:
        repository["forked_from"] = html.get("href")[1:]

    # Forks
    html = soup.select_one('[id="repo-network-counter"]')
    if html is not None:
        fork = html.get("title").replace(",", "")
        try:
            forks_count = int(fork)
            repository["forks_count"] = forks_count
        except ValueError:
            pass

    # Stars
    html = soup.select_one('[id="repo-stars-counter-star"]')
    if html is not None:
        star = html.get("title").replace(",", "")
        try:
            stargazers_count = int(star)
            repository["stargazers_count"] = stargazers_count
        except ValueError:
            pass

    # Issues
    html = soup.select_one('[id="issues-repo-tab-count"]')
    if html is not None:
        issues = html.get("title").replace(",", "")
        try:
            issues_count = int(issues)
            repository["issues_count"] = issues_count
        except ValueError:
            pass

    # Pull requests
    html = soup.select_one('[id="pull-requests-repo-tab-count"]')
    if html is not None:
        pull_requests = html.get("title").replace(",", "")
        try:
            pull_requests_count = int(pull_requests)
            repository["pull_requests_count"] = pull_requests_count
        except ValueError:
            pass

    # Description
    html = soup.select_one('[class="f4 my-3"]')
    if html is not None:
        repository["description"] = html.text.strip()

    # Web site
    html = soup.select_one('[role="link"]')
    if html is not None:
        repository["website"] = html.get("href")

    # License or "specific"
    html = soup.select('[class="sr-only"]')
    for item in html:
        if item.text.strip() == "License":
            html2 = item.next_sibling.next_sibling
            html3 = html2.select_one('a')
            repository["license"] = html3.text.strip()
            if repository["license"] == "View license":
                repository["license"] = "specific"
            break

    # Topics
    html = soup.select('[class="topic-tag topic-tag-link"]')
    if html is not None:
        topics = []
        for item in html:
            topics.append(item.get("title").split()[1])
        repository["topics"] = topics
        repository["topics_count"] = len(topics)

    # Watching
    html = soup.select('[class="Link Link--muted"]')
    for item in html:
        if item.get("href").endswith("/watchers"):
            html2 = item.select_one('strong')
            if html2 is not None:
                try:
                    repository["watching_count"] = int(html2.text.strip())
                except ValueError:
                    pass
            break

    # Releases and Packages
    html = soup.select('[class="BorderGrid-cell"]')
    for item in html:
        html2 = item.select_one('[class*="Link--primary"]')
        if html2 is not None:
            href = html2.get("href")
            html3 = html2.select_one('[class*="Counter"]')
            if href.endswith("/releases"):
                if html3 is not None:
                    try:
                        repository["releases_count"] = int(html3.text.strip().replace(",", ""))
                    except ValueError:
                        pass

                    # Last release name & date
                    html4 = item.select_one('[class*="css-truncate"]')
                    if html4 is not None:
                        repository["last_release_name"] = html4.text.strip()
                    html4 = item.select_one('relative-time')
                    if html4 is not None:
                        repository["last_release_date"] = html4.get("datetime")

                else:
                    repository["releases_count"] = 0
            elif "/packages" in href:
                if html3 is not None:
                    try:
                        repository["packages_count"] = int(html3.text.strip().replace(",", ""))
                    except ValueError:
                        pass

    # Is sponsorable?
    html = soup.select('h2[class="h4 mb-3"]')
    repository["is_sponsorable"] = False
    for item in html:
        text = item.text.strip()
        if text == "Sponsor this project":
            repository["is_sponsorable"] = True

    # Contributors
    html = soup.select_one('[class="Counter ml-1"]')
    if html is not None:
        contributors = html.get("title").replace(",", "")
        try:
            contributors_count = int(contributors)
            repository["contributors_count"] = contributors_count
        except ValueError:
            pass

    # Languages
    html = soup.select('[itemprop="keywords"]')
    if html is not None:
        languages = {}
        repository["programming_language"] = ""
        for item in html:
            value = item.get("aria-label").split()
            language = " ".join(value[:-1])
            percentage = value[-1]
            try:
                languages[language] = float(percentage)
            except ValueError:
                languages[language] = percentage
            if not repository["programming_language"]:
                repository["programming_language"] = language
        repository["languages"] = languages

    # Commits
    html = soup.select('span[class*="Text-sc-"]')
    for item in html:
        content = item.text.strip().replace(",", "")
        if content.endswith(" Commits"):
            try:
                repository["commits_count"] = int(content[:-8])
            except ValueError:
                pass
        elif content == "1 Commit":
            repository["commits_count"] = 1

    return repository

####################################################################################################
def load_repositories(repositories_list, cache_days, force_fetch=False):
    """ Returns a dictionary of repositories information """
    accounts = {}

    for item in repositories_list:
        if item.count('/') == 1:
            account = item.split('/')[0]
            repository = item.split('/')[1]

            if account not in accounts:
                accounts[account] = {"repositories":{}}

            accounts[account]["repositories"][repository] = load_repository(
                account,
                repository,
                cache_days,
                force_fetch=force_fetch
            )
        else:
            logging.error(
                "libgh: Repositories parameters must be in 'account/repo' form. '%s' discarded",
                item
            )

    return accounts



#!/usr/bin/env python3
""" libgh - GitHub scraping library
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import getopt
import json
import logging
import os
import pprint
import sys

import libpnu

from .common import CACHE_DIR, CACHE_DAYS
from .accounts import load_account, load_accounts
from .repositories import load_repository, load_repositories
from .libpnu2 import get_url_bdata, get_url_data, prune_cache, collection2xml

# Version string used by the what(1) and ident(1) commands:
ID = "@(#) $Id: libgh - GitHub scraping tool v0.9.4 (May 29, 2024) by Hubert Tournier $"

# Default parameters. Can be overcome by environment variables, then command line options
parameters = {
    "Prune cache": False,
    "Force fetching URL": False,
    "Cache days": CACHE_DAYS,
    "Complete partial": [],
    "JSON output": False,
    "XML output": False,
}


####################################################################################################
def _display_help():
    """ Display usage and help """
    #pylint: disable=C0301
    print("usage: libgh [--debug] [--help|-?] [--version]", file=sys.stderr)
    print("       [--from] [--json|-j] [--topics] [--xml|-x]", file=sys.stderr)
    print("       [--days|-d DAYS] [--force|-f] [--prune|-p]", file=sys.stderr)
    print("       [--] account_or_repo [...]", file=sys.stderr)
    print("  ------------------  --------------------------------------------------", file=sys.stderr)
    print("  --days|-d DAYS      Set number of caching days (0=don't use cache)", file=sys.stderr)
    print("  --force|-f          Force fetching URL instead of using cache", file=sys.stderr)
    print("  --from              Load repositories when forked_from is blank", file=sys.stderr)
    print("  --json|-j           Switch to JSON output instead of plain text", file=sys.stderr)
    print("  --prune|-p          Prune cache items olday than DAYS and cache index", file=sys.stderr)
    print("  --topics            Load repositories when there are missing topics", file=sys.stderr)
    print("  --xml|-x            Switch to XML output instead of plain text", file=sys.stderr)
    print("  --debug             Enable debug mode", file=sys.stderr)
    print("  --help|-?           Print usage and this help message and exit", file=sys.stderr)
    print("  --version           Print version and exit", file=sys.stderr)
    print("  --                  Options processing terminator", file=sys.stderr)
    print(file=sys.stderr)
    #pylint: enable=C0301


####################################################################################################
def _process_command_line():
    """ Process command line options """
    #pylint: disable=C0103, W0602
    global parameters
    #pylint: enable=C0103, W0602

    # option letters followed by : expect an argument
    # same for option strings followed by =
    character_options = "d:fjpx?"
    string_options = [
        "days=",
        "debug",
        "force",
        "from",
        "help",
        "json",
        "prune",
        "topics",
        "version",
        "xml",
    ]

    try:
        options, remaining_arguments = getopt.getopt(
            sys.argv[1:], character_options, string_options
        )
    except getopt.GetoptError as error:
        logging.critical("Syntax error: %s", error)
        _display_help()
        sys.exit(1)

    for option, argument in options:

        if option == "--debug":
            logging.disable(logging.NOTSET)

        elif option in ("--help", "-?"):
            _display_help()
            sys.exit(0)

        elif option == "--version":
            print(ID.replace("@(" + "#)" + " $" + "Id" + ": ", "").replace(" $", ""))
            sys.exit(0)

        elif option in ("--days", "-d"):
            try:
                parameters["Cache days"] = int(argument)
            except ValueError:
                logging.critical("--days|-d parameter is not an integer number")
                sys.exit(1)
            if parameters["Cache days"] < 0:
                logging.critical("--days|-d parameter is not a positive integer number")
                sys.exit(1)

        elif option in ("--force", "-f"):
            parameters["Force fetching URL"] = True

        elif option == "--from":
            parameters["Complete partial"].append("forked_from")

        elif option in ("--json", "-j"):
            parameters["JSON output"] = True
            parameters["XML output"] = False

        elif option in ("--prune", "-p"):
            parameters["Prune cache"] = True

        elif option == "--topics":
            parameters["Complete partial"].append("topics")

        elif option in ("--xml", "-x"):
            parameters["XML output"] = True
            parameters["JSON output"] = False

    return remaining_arguments


####################################################################################################
def main():
    """ The program's main entry point """
    program_name = os.path.basename(sys.argv[0])

    libpnu.initialize_debugging(program_name)
    libpnu.handle_interrupt_signals(libpnu.interrupt_handler_function)
    arguments = _process_command_line()

    if parameters["Prune cache"]:
        prune_cache(CACHE_DIR, parameters["Cache days"])

    if not arguments:
        _display_help()

    else:
        accounts = []
        repositories = []
        for argument in arguments:
            if '/' in argument:
                repositories.append(argument)
            else:
                accounts.append(argument)

        data = {}
        if accounts:
            data = load_accounts(
                accounts,
                parameters["Cache days"],
                force_fetch=parameters["Force fetching URL"],
                complete=parameters["Complete partial"],
            )
        if repositories:
            data2 = load_repositories(
                repositories,
                parameters["Cache days"],
                force_fetch=parameters["Force fetching URL"],
            )

            # Performing nested dictionary update
            for account, account_value in data2.items():
                if account in data:
                    for repository, repository_value in account_value["repositories"].items():
                        data[account]["repositories"][repository] = repository_value
                else:
                    data[account] = account_value

        if parameters["JSON output"]:
            json.dump(data, sys.stdout)
            print()
        elif parameters["XML output"]:
            xml = collection2xml(data, name="GitHub")
            for line in xml:
                print(line)
        else:
            pprint.pprint(data, sort_dicts=False)

    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
""" libgh - GitHub scraping library and tool
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging

from bs4 import BeautifulSoup

from .common import GITHUB_URL, CACHE_DIR, REQUESTS_MIN_DELAY, REQUESTS_PER_HOUR
from .libpnu2 import get_url_data
from .personal_account import load_user_account
from .organization_account import load_org_account

####################################################################################################
def load_account(account_name, cache_days, force_fetch=False, complete=[]):
    """ Returns a dictionary of account information """
    account = {}

    try:
        data, response = get_url_data(
            f"{GITHUB_URL}/{account_name}",
            CACHE_DIR,
            cache_days=cache_days,
            cache_index=True,
            force_fetch=force_fetch,
            min_delay=REQUESTS_MIN_DELAY,
            max_per_hour=REQUESTS_PER_HOUR
        )
    except (LookupError, PermissionError) as error:
        logging.error("libgh: %s", error)
        return account
    for item in response:
        if item[0].startswith("x-ratelimit"):
            logging.debug("libgh: HTTP response: %s=%s", item[0], item[1])

    soup = BeautifulSoup(data, "html.parser")

    organization = soup.find("meta", attrs={"name":"hovercard-subject-tag"})
    if organization is None:
        account = load_user_account(
            account_name,
            soup,
            cache_days,
            force_fetch=force_fetch,
            complete=complete
        )
    else:
        account = load_org_account(
            account_name,
            soup,
            cache_days,
            force_fetch=force_fetch,
            complete=complete
        )

    return account

####################################################################################################
def load_accounts(accounts_list, cache_days, force_fetch=False, complete=[]):
    """ Returns a dictionary of accounts information """
    accounts = {}

    for account_name in accounts_list:
        accounts[account_name] = load_account(
            account_name,
            cache_days,
            force_fetch=force_fetch,
            complete=complete
        )

    return accounts



#!/usr/bin/env python3
""" libgh - GitHub scraping library and tool
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import json
import logging

from bs4 import BeautifulSoup

from .common import GITHUB_URL, CACHE_DIR, REQUESTS_MIN_DELAY, REQUESTS_PER_HOUR
from .libpnu2 import get_url_data
from .repositories import load_repository

####################################################################################################
def load_org_repositories(
    account_name,
    cache_days,
    force_fetch=False,
    complete=[],
    repos_type="all"
):
    """ Returns a dictionary of user's repositories """
    repos = {}
    page = 1

    while True:
        try:
            data, response = get_url_data(
                f"{GITHUB_URL}/orgs/{account_name}/repositories?type={repos_type}&page={page}",
                CACHE_DIR,
                cache_days=cache_days,
                cache_index=True,
                force_fetch=force_fetch,
                min_delay=REQUESTS_MIN_DELAY,
                max_per_hour=REQUESTS_PER_HOUR
            )
        except (LookupError, PermissionError) as error:
            logging.error("libgh: %s", error)
            return repos
        for item in response:
            if item[0].startswith("x-ratelimit"):
                logging.debug("libgh: HTTP response: %s=%s", item[0], item[1])

        soup = BeautifulSoup(data, "html.parser")

        # repositories list
        json_data = soup.find(
            "script",
            attrs={"type":"application/json", "data-target":"react-app.embeddedData"}
        )
        if json_data is not None:
            j = json.loads(json_data.text)
        else:
            # We received JSON data instead of HTML!
            j = json.loads(data)

        for repository in j["payload"]["repositories"]:
            name = repository["name"]
            repos[name] = {}
            uncomplete = False

            # forked from
            if repository["isFork"]:
                repos[name]["forked_from"] = "" # The original location is not provided
                if "forked_from" in complete:
                    uncomplete = True

            # topics
            if "topicNames" in repository:
                repos[name]["topics"] = repository["topicNames"]
                repos[name]["topics_count"] = len(repository["topicNames"])
                if "topicsNotShown" in repository:
                    repos[name]["topics_count"] += repository["topicsNotShown"]
            else:
                repos[name]["topics"] = []
                if "topicsNotShown" in repository:
                    repos[name]["topics_count"] = repository["topicsNotShown"]
            if "topicsNotShown" in repository and repository["topicsNotShown"] and "topics" in complete:
                uncomplete = True

            if uncomplete:
                repos[name] = load_repository(account_name, name, cache_days, force_fetch)

            else:
                # description
                if repository["description"]:
                    repos[name]["description"] = repository["description"]

                # programming language
                if repository["primaryLanguage"] is not None:
                    repos[name]["programming_language"] = repository["primaryLanguage"]["name"]

                # stargazers
                repos[name]["stargazers"] = {}
                repos[name]["stargazers_count"] = repository["starsCount"]

                # forks
                repos[name]["forks"] = {}
                repos[name]["forks_count"] = repository["forksCount"]

                # issues
                repos[name]["issues_count"] = repository["issueCount"]

                # pull requests
                repos[name]["pull_requests_count"] = repository["pullRequestCount"]

                # license
                if repository["license"] is not None:
                    repos[name]["license"] = repository["license"]

            # last updated
            repos[name]["last_updated"] = repository["lastUpdated"]["timestamp"]

            # archived?
            if repos_type == "archived":
                repos[name]["archived"] = True

        # is there a next page?
        next_page = soup.select_one('[aria-label="Next Page"]')
        if next_page is not None:
            try:
                next_page.get("href")
            except AttributeError:
                break
        elif page == j["payload"]["pageCount"]:
            break

        page += 1

    return repos


####################################################################################################
def load_org_account(account_name, soup, cache_days, force_fetch=False, complete=[]):
    """ Returns a dictionary of organization account information """
    account = {}

    # the number of Repositories, Teams and (public) People is not available from the tabs pane

    # account type
    account["account_type"] = "organization"

    # avatar
    html = soup.select_one('[itemprop="image"]')
    if html is not None:
        account["avatar"] = html.get("src")

    # organization name
    html2 = html.parent.next_sibling.next_sibling.contents[1]
    if html2 is not None:
        text = html2.text.strip()
        if text:
            account["name"] = text

    # "biography": organization sub title
    html2 = html.parent.next_sibling.next_sibling.contents[3]
    if html2 is not None:
        text = html2.text.strip()
        if text:
            account["biography"] = text

    # followers
    account["followers"] = {}
    html = soup.select('[class*="Link--secondary"]')
    for item in html:
        href = item.get("href")
        if href is None:
            continue
        if "followers" in href:
            number = [x.strip() for x in item.text.split("\n") if x][0]
            try:
                account["followers_count"] = int(number)
            except ValueError:
                if number.endswith('k'):
                    account["followers_count"] = int(float(number[:-1]) * 1000)
                elif number.endswith('m'):
                    account["followers_count"] = int(float(number[:-1]) * 1000000)
                else:
                    account["followers_count"] = 0

    # location
    html = soup.select_one('[itemprop="location"]')
    if html is not None:
        account["location"] = html.text.strip()

    # profile website URL
    html = soup.select_one('[itemprop="url"]')
    if html is not None:
        account["website"] = html.text.strip()

    # social media
    account["social"] = {}
    html = soup.select('[class*="Link--primary"]')
    for item in html:
        ident = item.text.strip()
        href = item.get("href")
        if "facebook.com/" in href:
            account["social"]["facebook"] = {"id": ident, "url": href}
        elif href.startswith("https://www.instagram.com/"):
            account["social"]["instagram"] = {"id": ident, "url": href}
        elif href.startswith("https://www.linkedin.com/"):
            account["social"]["linkedin"] = {"id": ident, "url": href}
        elif href.startswith("https://mastodon.social/"):
            account["social"]["mastodon"] = {"id": ident, "url": href}
        elif href.startswith("https://mozilla.social/"):
            account["social"]["mozilla"] = {"id": ident, "url": href}
        elif href.startswith("https://twitter.com/"):
            account["social"]["x"] = {"id": ident, "url": href}
        elif href.startswith("https://www.youtube.com/"):
            account["social"]["youtube"] = {"id": ident, "url": href}
        elif not href.startswith("http"):
            pass
        else:
            if "url" in account["social"]:
                account["social"]["url"].append(href)
            else:
                account["social"]["url"] = [href]

    # repositories
    account["repositories"] = {}
    html = soup.select_one('[data-autosearch-results]')
    if html is not None:
        number = html.text.split()[3]
        try:
            account["repositories_count"] = int(number)
        except ValueError:
            pass
        account["repositories"] = load_org_repositories(
            account_name,
            cache_days,
            force_fetch=force_fetch,
            complete=complete
        )
        if len(account["repositories"]) != account["repositories_count"]:
            account["repositories"].update(
                load_org_repositories(
                    account_name,
                    cache_days,
                    force_fetch=force_fetch,
                    complete=complete,
                    repos_type="archived"
                )
            )
    else:
        account["repositories_count"] = 0

    # repositories stars (computed)
    stars = 0
    for repo in account["repositories"]:
        stars += account["repositories"][repo]["stargazers_count"]
    account["repositories_stars"] = stars

    # the details of "People", "Top languages" and "Most used topics" are not available
    # from the right pane

    # top languages (computed)
    top_languages = {}
    for repo in account["repositories"]:
        if "programming_language" in account["repositories"][repo]:
            language = account["repositories"][repo]["programming_language"]
            if language in top_languages:
                top_languages[language] += 1
            else:
                top_languages[language] = 1
    account["top_languages"] = dict(sorted(
        top_languages.items(),
        key=lambda item: item[1],
        reverse=True
    ))

    # most used topics (computed)
    most_used_topics = {}
    for repo in account["repositories"]:
        if "topics" in account["repositories"][repo]:
            topics = account["repositories"][repo]["topics"]
            for topic in topics:
                if topic in most_used_topics:
                    most_used_topics[topic] += 1
                else:
                    most_used_topics[topic] = 1
    account["most_used_topics"] = dict(sorted(
        most_used_topics.items(),
        key=lambda item: item[1],
        reverse=True
    ))

    if len(account["repositories"]) != account["repositories_count"]:
        logging.warning(
            "libgh: Loaded %d/%d repositories for account '%s'",
            len(account["repositories"]),
            account["repositories_count"],
            account_name
        )

    return account



#!/usr/bin/env python3
""" libgh - GitHub scraping library and tool
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

# GitHub address
GITHUB_URL = "https://github.com"

# Caching
CACHE_DIR = "libgh"
CACHE_DAYS = 7

# Rate limits for GitHub
# See https://docs.github.com
#    /fr/apps/creating-github-apps/registering-a-github-app/rate-limits-for-github-apps
REQUESTS_MIN_DELAY = 1 # second(s) between Web requests
REQUESTS_PER_HOUR = 60 # requests per hour



#!/usr/bin/env python3
""" libgh - GitHub scraping libraryaand tool
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import logging
import re

from bs4 import BeautifulSoup

from .common import GITHUB_URL, CACHE_DIR, REQUESTS_MIN_DELAY, REQUESTS_PER_HOUR
from .libpnu2 import get_url_data
from .repositories import load_repository

####################################################################################################
def load_user_repositories(user_name, url, page, cache_days, force_fetch=False, complete=[]):
    """ Returns a dictionary of user's repositories """
    repos = {}

    try:
        data, response = get_url_data(
            url,
            CACHE_DIR,
            cache_days=cache_days,
            cache_index=True,
            force_fetch=force_fetch,
            min_delay=REQUESTS_MIN_DELAY,
            max_per_hour=REQUESTS_PER_HOUR
        )
    except (LookupError, PermissionError) as error:
        logging.error("libgh: %s", error)
        return repos
    for item in response:
        if item[0].startswith("x-ratelimit"):
            logging.debug("libgh: HTTP response: %s=%s", item[0], item[1])

    soup = BeautifulSoup(data, "html.parser")

    # repositories list
    repos_list = soup.select_one('[id="user-repositories-list"]')
    items = repos_list.select("li")
    for item in items:
        # name
        html = item.select_one('[itemprop="name codeRepository"]')
        name = html.text.strip()
        repos[name] = {}

        # topics
        repos[name]["topics"] = []
        html = item.select('[data-octo-click="topic_click"]')
        if html is not None:
            for html2 in html:
                repos[name]["topics"].append(html2.text.strip())

        # when there are 7 topics, it's probable than GitHub is only showing the first ones
        # (can also happen with less when the topics character length is large)
        if "topics" in repos[name] and len(repos[name]["topics"]) == 7 and "topics" in complete:
            repos[name] = load_repository(user_name, name, cache_days, force_fetch)

        else:
            # archived?
            li_class = item.get("class")
            if "archived" in li_class:
                repos[name]["archived"] = True

            # forked from
            html = item.select_one('[class="Link--muted Link--inTextBlock"]')
            if html is not None:
                repos[name]["forked_from"] = html.text.strip()

            # description
            html = item.select_one('[itemprop="description"]')
            if html is not None:
                repos[name]["description"] = html.text.strip()

            # programming language
            html = item.select_one('[itemprop="programmingLanguage"]')
            if html is not None:
                repos[name]["programming_language"] = html.text.strip()

            # stargazers
            repos[name]["stargazers"] = {}
            html = item.select_one('[class="octicon octicon-star"]')
            if html is None:
                repos[name]["stargazers_count"] = 0
            else:
                html2 = html.next_sibling
                repos[name]["stargazers_count"] = int(html2.text.strip().replace(',', ''))

            # forks
            repos[name]["forks"] = {}
            html = item.select_one('[class="octicon octicon-repo-forked"]')
            if html is None:
                repos[name]["forks_count"] = 0
            else:
                html2 = html.next_sibling
                repos[name]["forks_count"] = int(html2.text.strip().replace(',', ''))

            # license
            html = item.select_one('[class*="octicon octicon-law"]')
            if html is not None:
                html2 = html.next_sibling
                repos[name]["license"] = html2.text.strip()

        # updated
        html = item.select_one('relative-time')
        if html is not None:
            repos[name]["last_updated"] = html.get("datetime")

    # is there a next page?
    next_page = soup.select_one('[class="next_page"]')
    try:
        href = next_page.get("href")
    except AttributeError:
        href = ""
    if href:
        repos.update(
            load_user_repositories(
                user_name,
                GITHUB_URL + href,
                page + 1,
                cache_days,
                force_fetch=force_fetch
            )
        )

    return repos


####################################################################################################
def load_user_account(account_name, soup, cache_days, force_fetch=False, complete=[]):
    """ Returns a dictionary of personal account information """
    account = {}

    # account type
    account["account_type"] = "personal"

    # avatar
    html = soup.select_one('[itemprop="image"]')
    if html is not None:
        account["avatar"] = html.get("href")

    # (full)name
    html = soup.select_one('[itemprop="name"]')
    if html is not None:
        account["name"] = html.text.strip()

    # pronouns: not provided while unauthenticated
    #html = soup.select_one('[itemprop="pronouns"]')

    # bio(graphy)
    html = soup.select_one('[class*="p-note user-profile-bio"]')
    if html is not None:
        html_text = html.text.strip()
        if html_text:
            account["biography"] = html_text

    # sponsor
    html = soup.select_one('[class*="octicon octicon-heart icon-sponsor"]')
    if html is not None:
        account["sponsor"] = True

    # followers & following
    account["followers"] = {}
    account["following"] = {}
    html = soup.select('[class*="Link--secondary"]')
    for item in html:
        href = item.get("href")
        if href is None:
            continue
        if "followers" in href:
            number = [x.strip() for x in item.text.split("\n") if x][0]
            try:
                account["followers_count"] = int(number)
            except ValueError:
                if number.endswith('k'):
                    account["followers_count"] = int(float(number[:-1]) * 1000)
                elif number.endswith('m'):
                    account["followers_count"] = int(float(number[:-1]) * 1000000)
                else:
                    account["followers_count"] = 0
        if "following" in href:
            number = [x.strip() for x in item.text.split("\n") if x][0]
            try:
                account["following_count"] = int(number)
            except ValueError:
                if number.endswith('k'):
                    account["following_count"] = int(float(number[:-1]) * 1000)
                elif number.endswith('m'):
                    account["following_count"] = int(float(number[:-1]) * 1000000)
                else:
                    account["following_count"] = 0

    # company
    html = soup.select_one('[class="p-org"]')
    if html is not None:
        account["company"] = html.text.strip()

    # location
    html = soup.select_one('[itemprop="homeLocation"]')
    if html is not None:
        account["location"] = html.text.strip()

    # (localtime and) timezone
    html = soup.select_one('[itemprop="localTime"]')
    if html is not None:
        timezone = html.text.strip().split('\n')[1].strip()
        if timezone.startswith('('):
            account["timezone"] = timezone[1:-1]
        else:
            account["timezone"] = timezone

    # public email: not provided while unauthenticated
    #html = soup.select_one('[itemprop="email"]')

    # profile website URL
    html = soup.select_one('[data-test-selector="profile-website-url"]')
    if html is not None:
        account["website"] = html.text.strip()

    # social media
    account["social"] = {}
    html = soup.select('[itemprop="social"]')
    for item in html:
        anchor = item.select_one('[class="Link--primary"]')
        if anchor is not None:
            ident = anchor.text.strip()
            href = anchor.get("href")
            if href.startswith("https://www.facebook.com/"):
                account["social"]["facebook"] = {"id": ident, "url": href}
            elif href.startswith("https://www.instagram.com/"):
                account["social"]["instagram"] = {"id": ident, "url": href}
            elif href.startswith("https://www.linkedin.com/"):
                account["social"]["linkedin"] = {"id": ident, "url": href}
            elif href.startswith("https://mastodon.social/"):
                account["social"]["mastodon"] = {"id": ident, "url": href}
            elif href.startswith("https://mozilla.social/"):
                account["social"]["mozilla"] = {"id": ident, "url": href}
            elif href.startswith("https://twitter.com/"):
                account["social"]["x"] = {"id": ident, "url": href}
            elif href.startswith("https://www.youtube.com/"):
                account["social"]["youtube"] = {"id": ident, "url": href}
            else:
                if "url" in account["social"]:
                    account["social"]["url"].append(href)
                else:
                    account["social"]["url"] = [href]

    # achievements
    account["achievements"] = {}
    html = soup.select('[class="position-relative"]')
    for item in html:
        href = item.get("href")
        if href is not None:
            if "?achievement=" in href:
                achievement = re.sub(r".*achievement=", "", href)
                achievement = re.sub(r"&.*", "", achievement)
                if achievement not in account["achievements"]:
                    account["achievements"][achievement] = {}

                    badge = item.select_one('[class="achievement-badge-sidebar"]')
                    if badge is not None:
                        account["achievements"][achievement]["badge"] = badge.get("src")

                    tier = item.select_one('[class*="Label achievement-tier-label"]')
                    count = 1
                    if tier is not None:
                        if tier.text.strip().startswith('x'):
                            count = int(tier.text.strip()[1:])
                    account["achievements"][achievement]["count"] = count

    # highlights
    account["highlights"] = []
    html = soup.select('[class="h4 mb-2"]')
    for item in html:
        if item.text.strip() == "Highlights":
            subitems = item.parent.select("li")
            for subitem in subitems:
                account["highlights"].append(subitem.text.strip())

    # organizations
    account["organizations"] = {}
    html = soup.select('[itemprop="follows"]')
    for item in html:
        org = item.get("href")[1:] # without the leading /
        account["organizations"][org] = {}

        img = item.select_one("img")
        if img is not None:
            avatar = img.get("src")
            account["organizations"][org]["avatar"] = avatar

    # repositories
    account["repositories"] = {}
    html = soup.select_one('[data-tab-item="repositories"]')
    if html is not None:
        try:
            account["repositories_count"] = int([x.strip() for x in html.text.split("\n") if x][-1])
        except ValueError:
            account["repositories_count"] = 0
        account["repositories"] = load_user_repositories(
            account_name,
            f"{GITHUB_URL}/{account_name}?tab=repositories",
            1,
            cache_days,
            force_fetch=force_fetch,
            complete=complete
        )
    else:
        account["repositories_count"] = 0

    # repositories stars (computed)
    stars = 0
    for repo in account["repositories"]:
        stars += account["repositories"][repo]["stargazers_count"]
    account["repositories_stars"] = stars

    # stars
    account["stars"] = {}
    html = soup.select_one('[data-tab-item="stars"]')
    if html is not None:
        try:
            account["stars_count"] = int([x.strip() for x in html.text.split("\n") if x][-1])
        except ValueError:
            account["stars_count"] = 0
    else:
        account["stars_count"] = 0

    # sponsoring
    html = soup.select_one('[data-tab-item="sponsoring"]')
    if html is not None:
        account["sponsoring"] = {}

    # yearly contributions: not included in what we load
    #html = soup.select_one('[class="js-yearly-contributions"]')

    # member since: not included in what we load
    # would be the last year in the timeline

    # top languages (computed)
    top_languages = {}
    for repo in account["repositories"]:
        if "programming_language" in account["repositories"][repo]:
            language = account["repositories"][repo]["programming_language"]
            if language in top_languages:
                top_languages[language] += 1
            else:
                top_languages[language] = 1
    account["top_languages"] = dict(sorted(
        top_languages.items(),
        key=lambda item: item[1],
        reverse=True
    ))

    # most used topics (computed)
    most_used_topics = {}
    for repo in account["repositories"]:
        if "topics" in account["repositories"][repo]:
            topics = account["repositories"][repo]["topics"]
            for topic in topics:
                if topic in most_used_topics:
                    most_used_topics[topic] += 1
                else:
                    most_used_topics[topic] = 1
    account["most_used_topics"] = dict(sorted(
        most_used_topics.items(),
        key=lambda item: item[1],
        reverse=True
    ))

    if len(account["repositories"]) != account["repositories_count"]:
        logging.warning(
            "libgh: Loaded %d/%d repositories for account '%s'",
            len(account["repositories"]),
            account["repositories_count"],
            account_name
        )

    return account



#!/usr/bin/env python3
""" libgh - GitHub scraping library and tool
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import datetime
import hashlib
import logging
import lzma
import os
import re
import time
import urllib.request

import libpnu

HTTP_HEADERS = {
    "User-Agent":
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:124.0) Gecko/20100101 Firefox/124.0",
    "Accept": "*/*",
    "Accept-Language": "en;q=1.0, *;q=0.5",
    "Accept-Encoding": "identity",
    "Connection": "close",
}


####################################################################################################
def _count_from(current_time, requests_times):
    """ Returns the number of requests made in the last day, hour and minute """
    last_day = 0
    last_hour = 0
    last_minute = 0

    one_day = 24 * 60 * 60 # seconds
    one_hour = 60 * 60 # seconds
    one_minute = 60 # seconds

    # The requests_times is not necessarily ordered by ascending timestamps...
    for request_time in requests_times:
        time_diff = current_time - request_time
        if time_diff <= one_day:
            last_day += 1
        if time_diff <= one_hour:
            last_hour += 1
        if time_diff <= one_minute:
            last_minute += 1

    return last_day, last_hour, last_minute

####################################################################################################
def get_url_bdata(
    url,
    cache_dir,
    headers={},
    cache_days=1,
    cache_index=False,
    force_fetch=False,
    min_delay=0,
    max_per_minute=-1,
    max_per_hour=-1,
    max_per_day=-1,
    timeout=15
):
    """ Returns an URL binary data and HTTP headers from a recent cache file or from the Web """
    current_time = time.time()
    directory = libpnu.get_caching_directory(cache_dir)

    # Keep a persistent history of URL fetched from the Web
    if not hasattr(get_url_bdata, "history"):
        get_url_bdata.history = {}

        # Load relevant history from the index file if it exist
        index_name = f"{directory}{os.sep}index.txt"
        lines = []
        if os.path.isfile(index_name):
            with open(index_name, encoding="utf-8", errors="ignore") as file:
                lines = file.read().splitlines()
        for line in lines:
            fields = line.split("#")
            date = datetime.datetime.strptime(fields[2], "%Y-%m-%d %H:%M:%S.%f")
            timestamp = datetime.datetime.timestamp(date)
            if current_time - timestamp <= cache_days * 24 * 60 * 60:
                website = re.sub(r"https*://", "", fields[0])
                website = re.sub(r"/.*", "", website)
                if website in get_url_bdata.history:
                    get_url_bdata.history[website].append(timestamp)
                else:
                    get_url_bdata.history[website] = [timestamp]

    # If there's a fragment reference in the URL, remove it
    # (else we would risk storing several time the same file)
    if '#' in url:
        url = re.sub(r"#.*", "", url)
    if "%23" in url:
        url = re.sub(r"%23.*", "", url)

    # Create an obfuscated and sanitized cache file name from the URL
    # (both to avoid any special characters in filenames and to offer
    # some privacy about what's in the cache, unless an index file is
    # used of course)
    digest = hashlib.sha256()
    digest.update(str.encode(url, encoding="utf-8", errors="ignore"))
    cache_file = digest.hexdigest() + ".xz"

    # Add 2 levels of sub directories to avoid using a flat storage space
    # (for 65K maximum sub directories)
    cache_file = re.sub(r"^(..)(..)(.*)", f"\g<1>{os.sep}\g<2>{os.sep}\g<3>", cache_file)

    # Make it a full pathname
    if directory:
        filename = directory + os.sep + cache_file
    else:
        filename = cache_file

    # Create directories if needed
    os.makedirs(os.path.dirname(filename), exist_ok=True)

    # Use a cache file if cache_days != 0
    if not force_fetch and cache_days:
        # If there's a cache file of less than cache_days, return its data
        if os.path.isfile(filename) \
        and (current_time - os.path.getmtime(filename)) < cache_days * 24 * 60 * 60:
            logging.debug("libpnu/get_url_bdata: Loading '%s' from cache", url)
            with lzma.open(filename) as file:
                bdata = file.read()
            return bdata, []

    # Else fetch the URL contents from the Web

    # Shall we wait before fetching the URL?
    current_date = datetime.datetime.fromtimestamp(current_time)
    website = re.sub(r"https*://", "", url)
    website = re.sub(r"/.*", "", website)
    if website in get_url_bdata.history:
        last_day, last_hour, last_minute = _count_from(current_time, get_url_bdata.history[website])
        logging.debug(
            "libpnu/get_url_bdata: Web requests to '%s' last day=%d, hour=%d, min=%d",
            website,
            last_day,
            last_hour,
            last_minute
        )
        slowing_down = False
        if last_day >= max_per_day > 0:
            logging.debug("libpnu/get_url_bdata: Max requests per day reached. Sleeping for 1 day")
            time.sleep(24 * 60 * 60)
            slowing_down = True
        if last_hour >= max_per_hour > 0:
            logging.debug(
                "libpnu/get_url_bdata: Max requests per hour reached. Sleeping for 1 hour"
            )
            time.sleep(60 * 60)
            slowing_down = True
        if last_minute >= max_per_minute > 0:
            logging.debug(
                "libpnu/get_url_bdata: Max requests per minute reached. Sleeping for 1 minute"
            )
            time.sleep(60)
            slowing_down = True
        if slowing_down:
            current_time = time.time()
            current_date = datetime.datetime.fromtimestamp(current_time)

        if min_delay:
            time_diff = current_time - get_url_bdata.history[website][-1]
            if time_diff < min_delay:
                time_wait = min_delay - time_diff
                logging.debug(
                    "libpnu/get_url_bdata: Slowing down URL fetching. Sleeping for %f seconds",
                    time_diff
                )
                time.sleep(time_wait)
                current_time = time.time()
                current_date = datetime.datetime.fromtimestamp(current_time)

        get_url_bdata.history[website].append(current_time)
    else:
        get_url_bdata.history[website] = [current_time]

    # Let's go now!
    logging.debug("libpnu/get_url_bdata: Loading '%s' from the Web", url)
    all_headers = HTTP_HEADERS
    for key, value in headers.items():
        all_headers[key] = value
    request = urllib.request.Request(url, headers=all_headers)
    response = []
    try:
        with urllib.request.urlopen(request, timeout=timeout) as http:
            bdata = http.read()
        response = http.getheaders()
    except urllib.error.HTTPError as error:
        if error == 'HTTP Error 429: Too Many Requests':
            raise PermissionError(f"Max requests reached for '{website}'") from error

        if error == 'HTTP Error 404: Not Found':
            # Write an empty file to avoid retrying later...
            with lzma.open(filename, "wb") as file:
                pass

        raise LookupError(f"Error while fetching '{url}': {error}") from error

    # Cache the URL contents for future requests
    with lzma.open(filename, "wb") as file:
        file.write(bdata)

    # If requested, write the correspondance between the url and the obfuscated cache file name
    if cache_index:
        if directory:
            index_filename = f"{directory}{os.sep}index.txt"
        else:
            index_filename = "index.txt"
        with open(index_filename, "at", encoding="utf-8", errors="ignore") as file:
            file.write(f"{url}#{cache_file}#{current_date}\n")

    return bdata, response

####################################################################################################
def get_url_data(
    url,
    cache_dir,
    headers={},
    cache_days=1,
    cache_index=False,
    force_fetch=False,
    min_delay=0,
    max_per_minute=-1,
    max_per_hour=-1,
    max_per_day=-1,
    timeout=15
):
    """ Returns an URL text data and HTTP headers from a recent cache file or from the Web """
    bdata, response = get_url_bdata(
        url,
        cache_dir,
        headers=headers,
        cache_days=cache_days,
        cache_index=cache_index,
        force_fetch=force_fetch,
        min_delay=min_delay,
        max_per_minute=max_per_minute,
        max_per_hour=max_per_hour,
        max_per_day=max_per_day,
        timeout=timeout
    )

    return bdata.decode("utf-8", "ignore"), response

####################################################################################################
def prune_cache(cache_dir, cache_days):
    """ Removes old files and their index entries, and unused directories from the cache dir """
    directory = libpnu.get_caching_directory(cache_dir)
    if not directory:
        # Don't take the risk to remove files
        return

    # Load the index file if it exist
    index_name = f"{directory}{os.sep}index.txt"
    lines = []
    if os.path.isfile(index_name):
        with open(index_name, encoding="utf-8", errors="ignore") as file:
            lines = file.read().splitlines()

    # Remove duplicates
    index = {}
    for line in lines:
        fields = line.split("#")
        url = fields[0]
        filename = fields[1]
        timestamp = fields[2]
        index[filename] = {"url": url, "timestamp": timestamp}

    # Remove old files and their index entries
    current_time = time.time()
    for root, _, files in os.walk(directory):
        for file in files:
            if file != "index.txt":
                pathname = root + os.sep + file
                filename = re.sub(f"^{directory}{os.sep}", "", pathname)
                if current_time - os.path.getmtime(pathname) > cache_days * 24 * 60 * 60:
                    os.remove(pathname)
                    if filename in index:
                        del index[filename]

    # Recreate the index file if needed
    if index:
        with open(index_name, "w", encoding="utf-8", errors="ignore") as file:
            for cache_file, value in index.items():
                url = value["url"]
                timestamp = value["timestamp"]
                file.write(f"{url}#{cache_file}#{timestamp}\n")

    # Remove empty directories
    subdirs = list(os.walk(directory, topdown=False))
    for subdir in subdirs:
        try:
            os.rmdir(subdir[0])
        except OSError:
            # subdir is not empty
            pass

####################################################################################################
def collection2xml(data, name="root", depth=0, indent_size=2, show_types=False):
    """ Returns the collection data type as a list of XML lines """
    xml = []
    data_type = str(type(data))[8:-2]

    if depth == 0:
        xml.append('<?xml version="1.0" encoding="UTF-8" ?>')
        xml.append(f"<{name}>")
        xml += collection2xml(
            data,
            depth=1,
            indent_size=indent_size,
            show_types=show_types
        )
        xml.append(f"</{name}>")
    elif data_type == "dict":
        for key, value in data.items():
            if key.lower().startswith("xml") \
            or not key[0].isalpha() and key[0] != '_':
                key = f"_{key}"
            key = key.replace(" ", "_")
            key = key.replace("/", "&sol;")
            value_type = str(type(value))[8:-2]
            strtype = ""
            if show_types:
                strtype = f' type="{value_type}"'
            if value_type == "dict":
                xml.append(f"{depth * indent_size * ' '}<{key}{strtype}>")
                xml += collection2xml(
                    value,
                    depth=depth + 1,
                    indent_size=indent_size,
                    show_types=show_types
                )
                xml.append(f"{depth * indent_size * ' '}</{key}>")
            elif value_type == "list":
                xml.append(f"{depth * indent_size * ' '}<{key}{strtype}>")
                xml += collection2xml(
                    value,
                    name=key,
                    depth=depth + 1,
                    indent_size=indent_size,
                    show_types=show_types
                )
                xml.append(f"{depth * indent_size * ' '}</{key}>")
            elif value_type in ("bool", "int", "float", "complex"):
                xml.append(f"{depth * indent_size * ' '}<{key}{strtype}>{value}</{key}>")
            elif value_type == "str":
                value = value.replace("&","&amp;")
                value = value.replace(">","&gt;")
                value = value.replace("<","&lt;")
                xml.append(f"{depth * indent_size * ' '}<{key}{strtype}>{value}</{key}>")
            else:
                logging.error("libpnu/collection2xml: value type '%s' not processed. Skipping it")
    elif data_type in ("tuple", "list", "set", "frozenset"):
        for value in data:
            value_type = str(type(value))[8:-2]
            strtype = ""
            if show_types:
                strtype = f' type="{value_type}"'
            if value_type == "dict":
                xml.append(f"{depth * indent_size * ' '}<{name}_item{strtype}>")
                xml += collection2xml(
                    value,
                    depth=depth + 1,
                    indent_size=indent_size,
                    show_types=show_types
                )
                xml.append(f"{depth * indent_size * ' '}</{name}_item>")
            elif value_type == "list":
                xml.append(f"{depth * indent_size * ' '}<{name}_item{strtype}>")
                xml += collection2xml(
                    value,
                    depth=depth + 1,
                    indent_size=indent_size,
                    show_types=show_types
                )
                xml.append(f"{depth * indent_size * ' '}</{name}_item>")
            elif value_type in ("bool", "int", "float", "complex"):
                xml.append(
                    f"{depth * indent_size * ' '}<{name}_item{strtype}>{value}</{name}_item>"
                )
            elif value_type == "str":
                value = value.replace("&","&amp;")
                value = value.replace(">","&gt;")
                value = value.replace("<","&lt;")
                xml.append(
                    f"{depth * indent_size * ' '}<{name}_item{strtype}>{value}</{name}_item>"
                )
            else:
                logging.error("libpnu/collection2xml: value type '%s' not processed. Skipping it")

    return xml
